#include "chess.h"

HashKey hashRnd[2][8][64];

typedef struct{
  HashKey key;  // for compare position
  HashKey key_side[2];
  int mtl[2];   // |
  int ext_key;  // castling, en_pass enable
  
  int score,flag,depth;
  Move mv;
}HItem;

int Hash_Size = (1<<19);


HItem *hash[2];


static unsigned long seed16=11111;  // start value (srand())
 
int rand16(void) {

 seed16=seed16*1103515245+12345;
 return((unsigned int)(seed16/65536)%32768);
}

void HashClear(void){
    memset(hash[WHITE],0,sizeof(HItem) * Hash_Size);
    memset(hash[BLACK],0,sizeof(HItem) * Hash_Size);
}

void HashDepthLow(void){
  int c,j,max_depth;
  do{
   max_depth = 0;
   for(c=WHITE; c <= BLACK; c++)
   {
     for(j = 0; j < Hash_Size;j++)
     {
       max_depth = max(max_depth, hash[c][j].depth /= 2);
     }
   }
  }while(max_depth > 2);
}

U64 Rand64(void){

 return      ((U64)rand16() << 60) ^
             ((U64)rand16() << 45) ^
             ((U64)rand16() << 30) ^
             ((U64)rand16() << 15) ^
             ((U64)rand16());
}


void HashInit(void){

    extern char* property_get(char name[]);

    int c, p, sq, tmp;
    
    printf("# hash table init: \n" );


    // get size 
    
    tmp = 1 << (atoi(property_get("hash size"))/2);
    if( tmp >= (1<<16) )
      Hash_Size = tmp;
    else{
      fprintf(stderr,"# error: \"hash size\"=%s   \n",
                     property_get("hash size"));
      exit(1);
    }


    // init
    
    for(c = WHITE; c <= BLACK; c++){
       hash[c] = (HItem*)malloc(sizeof(HItem)*Hash_Size);
       if(hash[c]==NULL)
       {
          fprintf(stderr,"# error: HashInit: out of memory \n");
          exit(1);
       }
       for(p = PAWN; p <= KING; p++)
        for(sq = 0; sq < 64; sq++)
          hashRnd[c][p][sq] = Rand64();

    }
    
    printf("# get mem %d byte(s) \n",
           sizeof(HItem)*Hash_Size*2 ); 
}



int HashLook(int alpha,int beta, int depth, int *ret_score, Move *ret_mv){

   int pos_ext_key(void);

   HItem *p = (HItem*)&hash[g.side][(int)g.key & (Hash_Size-1)];

   if(     p->key        == g.key )
    if(    p->key_side[0]== g.key_side[0])
     if(   p->key_side[1]== g.key_side[1])
      if(  p->mtl[WHITE] == g.mtl[WHITE])
       if( p->mtl[BLACK] == g.mtl[BLACK])
        if(p->ext_key    == pos_ext_key())
   {
      *ret_mv = p->mv;
      if(p->depth >= depth)
        switch(p->flag){
          case ALPHA:
             if(p->score <= alpha){
               *ret_score = p->score;
               return 1;
             }
          break;
          case EXACT:

               *ret_score = p->score;
               return 1;


          case BETA:
             if(p->score >= beta){
               *ret_score = p->score;
               return 1;
             }
          break;
        }

   }

   return 0;
}

int HashEmptyNode(int c, HashKey key, int depth){
 HItem *p = (HItem*)&hash[c][(int)key & (Hash_Size-1)];
 return p->key != key && p->depth <= depth;

}



void HashInsert(int alpha, int beta, int depth, Move mv){

 int pos_ext_key(void);

 HItem *p = (HItem*)&hash[g.side][(int)g.key & (Hash_Size-1)];
 int flag;

 if(mv){
  flag = alpha < beta ? EXACT : BETA;
 }else flag  = ALPHA;

 if( //p->depth <= depth
    p->flag == EMPTY  ||
    p->flag*2 + p->depth <= flag*2 + depth
   // (p->flag==EMPTY || p->depth <= depth || (p->flag == ALPHA && flag != ALPHA))
   ){
    p->score  = alpha;
    p->depth  = depth;
    p->key    = g.key;
    p->key_side[0] = g.key_side[0];
    p->key_side[1] = g.key_side[1];
    p->mtl[0] = g.mtl[0];
    p->mtl[1] = g.mtl[1];
    p->ext_key = pos_ext_key();
    p->flag   = flag;
    p->mv     = mv;
    
 }
}


int HashDamp(void){
  int c,j,full=0;

  for(c = WHITE; c <= BLACK; c++)
   for(j = 0; j < Hash_Size; j++)
     if(hash[c][j].depth)
       full++;

   return  full * 100 / (Hash_Size*2);
}
