#include "chess.h"
#include "pvtree.h"


/* init or clear */

void pvReset(PVTree *p) 
{ 

  memset(p,0,sizeof(PVTree)); 
  
}



/* alloc mem for node */

PVNode *pvNew(PVTree *p) 
{ 

   return p->pool_cnt < PV_MAX_NODES ? 
          &p->pool[p->pool_cnt++] : NULL; 
   
}



/*

  Вставляет строку перемещений в дерево от корня.
  Если некоторые узлы уже есть в дереве, то они
  просто передвигаются наверх.
  Если улов нет - они добавляются
  
*/

void pvInsertLine(PVTree *pv_tree, Move ret_pv_line[])
{
              int j;
              PVNode **p = &pv_tree->root;
              
              for( j = 0; j < 80 &&  ret_pv_line[j]; j++)
              {
                 PVNode **fn = pvFindMove(p, ret_pv_line[j]);
                 
                 if( fn ){
                    pvToFront(fn, p);
                 }else{
                 
                    pvInsertToFront(pv_tree, p,
                                    pvNewNode(pv_tree,
                                              ret_pv_line[j])
                                   );
                 }
                 p = &(*p)->nextPly;
              }
}



/*  поиск хода в одном списке (из одной позиции)  */
PVNode **pvFindMove(PVNode **list, move mv){
   if( list )
    while( *list ){
      if( (*list)->mv == mv )
          return list;
      list = &(*list)->moves;
    }
   return NULL;
}




/*
  переставляет узел p наверх в списке moves, 
  first - начало списка
*/

PVNode  **pvToFront(PVNode **p, PVNode **first){

  if( p && first && (*p) != (*first) ){
    PVNode *tmp = *p;
    
      
    *p = tmp->moves;  //исключен
    tmp->moves = *first;
    *first = tmp;     //включен в начало
  }
  return first;
}



/*
  новый узел.
  Все поля изначально уже обнулены.
*/
PVNode *pvNewNode(PVTree *tr, move mv){

   if(tr)
    if(tr->pool_cnt < PV_MAX_NODES){
      PVNode  *p = &tr->pool[ tr->pool_cnt++ ];

      p->mv = mv;
      return p;
    }
   return NULL;
}




/*
  Вставляет новый узел p в начало списка first.
 
*/
PVNode **pvInsertToFront( PVTree* tr, PVNode** first, PVNode* p){

  if( first && p && (*first) != p){
    PVNode *tmp;
    
    p->moves = *first; 
    *first = p;
  }
  return first;
}




/*
  вставляет новый узел в конец списка
*/
PVNode **pvInsertToEnd( PVTree* tr, PVNode** first, PVNode* p){

  if( first && p && (*first) != p){
     PVNode **list = first;
     int find = 0;
     
     while( *list ){
       list = &(*list)->moves;
     }   
     
     p->moves = *list;
     *list = p;

     
     return list;
  }

  return first;
}

/*
  ИЩет точное совпадение по проигранной строке
  Возвращает начало списка для данной позиции.
  Если PVTree используется только в функции поиска,
  то вызов может быть следующим:

  p = pvFindNodeTree(tree, &game_list[game_cnt], ply);

     ply от 0

*/

PVNode **pvFindNodeTree( PVTree* tr, move game_list[], int cnt )
{
  int j;
  PVNode **p = &tr->root;
  if(*p){
   if(cnt == 0) return  p;
   for( j = 0; j < cnt; j++ ){
     if( (p = pvFindMove(p, game_list[j])) == NULL )
         return NULL;
     p = &(*p)->nextPly;
   }
  }
  return p;

}














/* рекурсивная печать дерева в файл */

#define SIZE_X 10
#define SIZE_Y 90
PVNode *pvTable[SIZE_Y][SIZE_X];

void pvPrint(PVNode *p, FILE *f){
  char* PrintMove( char buf[], move mv, int check );
  void pvTreeToTable(PVNode *p, int x, int y);
  int x,y;
  char s[16];

  memset( pvTable, 0, sizeof(pvTable));
  pvTreeToTable( p, 0, 0);

  for( y = 0; y < SIZE_Y; y++){
    int find = 0;

    for( x = 0; x < SIZE_X; x++ ){
      if( pvTable[y][x] ){
         find = 1;
         break;
      }   
    }
    if(find == 0) break;
    for( x = 0; x < SIZE_X; x++ ){
      if( pvTable[y][x] )
        fprintf( f, "%-7s",  PrintMove(s, pvTable[y][x]->mv, 0));
      else
        fprintf( f, "%-7s","-");
    }
    fprintf(f,"\n");
  }
}

/*
  Дерево в таблицу.


*/
void pvTreeToTable(PVNode *p, int x, int y){
   int GetFirstY(void);                 // весь узел печатается на одном X
   int cnt = 0;
   
   while( p  &&  cnt < 3){
     if(x < SIZE_X  && y < SIZE_Y){
       pvTable[y][x] = p;               //печатаем первый ход по старому Y
     }
     pvTreeToTable(p->nextPly, x + 1, y); // рекурсия x + 1, y
     p = p->moves;                        //следующий ход
     y = GetFirstY();                     //определим первый свободный Y
     cnt++;
   }
}

int GetFirstY(void){
  int x,y;
   for(y = SIZE_Y-1; y >= 0; y--)
    for(x = SIZE_X-1; x >= 0; x--)
     if( pvTable[y][x] ) return y + 1;
  return 0;
}


char* PrintMove( char buf[], move mv, int check ){

   char *MoveToStr(Move mv, char* str);

   MoveToStr(mv, buf);
   if(check) strcat(buf,"+");
   return buf;
}
