#ifndef pvtree_includeu
#define pvtree_includeu

#include "chess.h"

/*

  TREE OF PV-LINES

  Дерево глвных вариантов.
  

      Позиция  -> Следущая позиция
         \
         Позиция
         ...
         ...       <--- список ходов  (Позиций)
         ...
         Позиция
         

*/

typedef HashKey HKey;
typedef Move move;



/* type */


struct PVNode{           /* node of tree (position) */

  move mv;                   /* best move */


  struct PVNode *moves,      /* list of position from one position */
                 *nextPly;   /* position after current move */
  
};

typedef struct PVNode  PVNode;

#define PV_MAX_NODES 10000
#define PV_MAX_HASH  (1<<10)





/* type */

struct PVTree{

  PVNode pool[PV_MAX_NODES];   /* allocator */
  int    pool_cnt;
  
  PVNode *root;                /* tree */
  
};
typedef struct PVTree PVTree;








/* functions */

/* main */
void     pvReset         (PVTree *tr);             /* set all default */
void     pvInsertLine    (PVTree *pv_tree, Move ret_pv_line[]); /* like pv-table */
          
PVNode **pvFindNodeTree  /* search games line in tree, return list of next moves */
                         (PVTree *tr,    
                          move game_list[],   /* line for search: start = tree_root */
                          int ply             /* length of line   (tree_root..ply) */ 
                          ); 



/* alloc mem (private) */
PVNode  *pvNew           (PVTree *tr);             /* get mem */
PVNode  *pvNewNode       (PVTree *tr,    move mv); /* get mem and init */


/* search and manipulation */
PVNode **pvFindMove      (PVNode **list, move mv);
PVNode **pvToFront       (PVNode **p,    PVNode **first);
PVNode **pvInsertToFront (PVTree *tr,    PVNode **first,   PVNode *p);
PVNode **pvInsertToEnd   (PVTree *tr,    PVNode **first,   PVNode *p);


/* debug */
void     pvPrint         (PVNode *p,     FILE *f);





#endif

